;;; scim-bridge-ja.el

;; Copyright (C) 2008 S. Irie

;; Author: S. Irie
;; Maintainer: S. Irie
;; Keywords: Input Method, i18n

(defconst scim-bridge-ja-version "0.6.1")

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or
;; (at your option) any later version.

;; It is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
;; License for more details.

;; You should have received a copy of the GNU General Public
;; License along with this program; if not, write to the Free
;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
;; MA 02111-1307 USA

;;; Commentary:

;; このプログラムは、Emacs のための scim-bridge クライアントである
;; scim-bdidge.el に追加することにより、変数や関数の説明文字列を日本語の
;; 文章に差し替えます。これは公式の scim-bridge の一部ではありません。

;; SCIM (Smart Common Input Method) は 30以上の言語 (CJK と多くの
;; ヨーロッパ言語) をサポートする POSIX スタイルのオペレーティングシステム
;; (Linux や BSD) のためのインプットメソッド (IM) プラットフォームです。

;;
;; Installation:
;;
;; 初めに、このファイル (scim-bridge-ja.el) 及び scim-bridge.el を
;; パスの通ったディレクトリに保存し、scim-bridge.el をバイトコンパイル
;; して下さい。
;;
;; 以下の行を .emacs ファイルに追加します。
;;
;;   (require 'scim-bridge-ja)
;;
;; Emacs を起動するときは、コマンドラインから以下のように打ち込んで下さい。
;;
;;   XMODIFIERS=@im=none emacs
;;
;; 起動したら、scim-mode を ON にします。
;;
;;   M-x scim-mode
;;
;;
;; .emacs ファイルの設定例です。
;;
;;   (require 'scim-bridge)
;;   ;; 入力のモードをバッファごとに切り替える
;;   (setq scim-mode-local t)
;;   ;; C-SPC は Set Mark に使う
;;   (scim-define-common-key ?\C-\  nil)
;;   ;; C-/ は Undo に使う
;;   (scim-define-common-key ?\C-/ nil)
;;   ;; 変換候補ウインドウの横方向の位置を微調整する
;;   (setq scim-adjust-window-x-position 'gnome)
;;   ;; 起動時に scim-mode を自動的に ON にする
;;   (scim-mode t)
;;
;; 注意：このプログラムは GNU Emacs 22 以降のバージョンで動作します。
;; また、端末エミュレータ上では動作しません。
;;

;; History:
;; 2008-05-20  S. Irie
;;         * Add/modify translations
;;         * Version 0.6.1
;;
;; 2008-05-18  S. Irie
;;         * Experimental first release
;;         * Version 0.6.0
;;
;; 2008-05-15  S. Irie
;;         * Version 0.5.0

;; ToDo:

;;; Code:

(require 'scim-bridge)

(put 'scim
     'group-documentation
     "洗練された多言語入力プラットフォーム")

(put 'scim-mode
     'variable-documentation
  "scim-mode の ON/OFF を切り替えます。
この変数に値を直接代入しても正しい動作はしませんので、
 \\[customize] あるいは関数 `scim-mode' を用いてください。")

(put 'scim-basic
     'group-documentation
     "基本的な入力方法の設定（モードの切り替え、キーボードなど）")

(put 'scim-mode-local
     'variable-documentation
     "この値が nil 以外ならば、入力のモードを各バッファごとに切り替えること
ができるように、IMコンテクストが各々のバッファについて登録されます。")

(put 'scim-common-function-key-list
     'variable-documentation
     "このリストは、直接入力時及びプリエディット時において SCIM がどのキー
イベントを受けとるかを表します。このリストへの要素の追加・削除は、関数
`scim-define-common-key' を用いても行うことができます。
注意：このオプションでは ESC や C-x などのプレフィックスキーを設定しないで
ください。もし設定すると、Emacs が操作不能になるかもしれません。")

(put 'scim-preedit-function-key-list
     'variable-documentation
     "このリスト、はプリエディット領域が存在する時に SCIM がどのキーイベント
受けとるかを表します。このリストへの要素の追加・削除は、関数
`scim-define-preedit-key' を用いても行うことができます。")

(put 'scim-use-kana-ro-key
     'variable-documentation
     "もし jp-106 キーボードを用いて日本語のかな入力方式を行うなら、この
オプションをONにして、シフトキーを押下することなく仮名文字の「ろ」を入力
できるようにしてください。
　このオプションはXウインドウシステムの設定を `xmodmap' コマンドを呼び出す
ことにより書き換えます。したがって、他のすべてのアプリケーションに影響を
及ぼします。")

(put 'scim-key-release-delay
     'variable-documentation
     "もし jp-106 キーボードを用いて SCIM-Anthy で日本語の親指シフト入力方式
を行うなら、SCIM に「キーを離した」という情報を送るまでの遅延時間をこのオプ
ションで設定してください（単位は秒）。値は SCIM-Anthy の同時打鍵時間の設定値
以下にセットしてください。")
(put 'scim-key-release-delay
     'custom-type
     '(choice (const :tag "なし" nil)
	      (number :tag "遅延時間（秒）"
		      :value 0.1)))

(put 'scim-undo-by-committed-string
     'variable-documentation
     "この値が nil ならば、アンドゥを行うときに短い確定文字列が20文字を
越えない範囲で1つに連結され、逆に長い確定文字列は20文字ごとに分割されます。
値が nil 以外ならば、アンドゥは確定した文字列ごとに行われます。")

(put 'scim-clear-preedit-when-unexpected-event
     'variable-documentation
     "このオプションは、プリエディット中に予期せぬイベントが発生した場合
の振る舞いを指定します。例えば、マウスによる文字列の貼り付けを行おうとした
時などです。この値が nil 以外ならば、プリエディット領域は消去されます。")

;; Appearance
(put 'scim-appearance
     'group-documentation
  "外観の設定（フェイス、変換候補ウインドウなど）")

(put 'scim-preedit-default-face
  'face-documentation
  "このフェイスはプリエディット領域の全体を表します。")

(put 'scim-preedit-underline-face
  'face-documentation
  "「下線」で使われるフェイスです。")

(put 'scim-preedit-highlight-face
  'face-documentation
  "「強調」で使われるフェイスです。")

(put 'scim-preedit-reverse-face
  'face-documentation
  "「反転」で使われるフェイスです。")

(put 'scim-adjust-window-x-position
     'variable-documentation
     "このオプションは、変換候補ウインドウの位置をウインドウ内の変換候補と
行内の変換候補が縦方向に揃うように調整するかどうかということをを指定します。
この値が `gnome' ならば、GNOME のフォントサイズの設定値を用いて調整します。
一方、値が整数で与えられていれば、ウインドウをずらすピクセル数を表します。
　このオプションは、候補ウインドウが常時表示されるようなタイプの入力方法、
例えば SCIM-pinyin（中国語）等には適していません。なぜなら、入力位置が
画面の下の方にある時には候補ウインドウがカーソルを覆い隠してしまう事がある
からです。")
(put 'scim-adjust-window-x-position
     'custom-type
     '(choice (const :tag "GNOME のフォント設定を利用" gnome)
	      (integer :tag "ピクセル数を指定" :value 24)
	      (const :tag "off" nil)))

(put 'scim-adjust-window-y-position
     'variable-documentation
     "このオプションは、変換候補ウインドウの縦方向の表示位置を `xwininfo'
コマンドを利用して調節するかどうかを指定します。もし値が nil ならば、位置の
調整は行われず、ウインドウが少しだけずれる場合があります。")

(put 'scim-prediction-window-position
     'variable-documentation
     "（日本語入力時のみ）この変数の値は (POS . ADJ) というコンスセル
の形で与えます。もし POS の値が が nil 以外ならば、予測候補ウインドウが
プリエディット領域の先頭部分の下に表示されます。ADJ の値が nil 以外ならば、
横方向の位置が変換候補ウインドウと同じように微調整されます。")
(put 'scim-prediction-window-position
     'custom-type
     '(cons
       (choice :tag "位置（POS)"
	       (const :tag "プリエディット領域の末尾" nil)
	       (const :tag "プリエディット領域の先頭" t))
       (choice :tag "微調整（ADJ）"
	       (const :tag "変換候補ウインドウと同様にする" t)
	       (const :tag "off" nil))))


(put 'scim-mode-line-string
     'variable-documentation
     "この変数はマイナーモードを許可するかどうかを制御します。先頭の文字が
空白である値は短い文字列で、モードラインに表示されます。この文字列は、複数の
モードの文字列が同時に表示できるように、短くしなければなりません。")

;; Advanced settings
(put 'scim-expert
     'group-documentation
     "高度な設定")

(put 'scim-focus-update-interval
     'variable-documentation
     "ウインドウのフォーカスを監視する周期を秒単位で指定します。この設定を
有効にするには、scim-mode を再起動する必要があります。")

(put 'scim-bridge-timeout
     'variable-documentation
     "特別な理由がない限り、この設定は変更しないでください。")

(put 'scim-bridge-wait-reply
     'variable-documentation
     "特別な理由がない限り、この設定は変更しないでください。")

(put 'scim-bridge-socket-use-script
     'variable-documentation
     "この値が nil 以外ならば、UNIXドメインソケットを通じて scim-bridge
に接続するために、関数 `make-network-process' を用いる代わりにスクリプトを
使用することを強制します。Emacs のバージョンが22以降の場合、普通はこのスクリ
プトを使う必要はありません。")

(put 'scim-bridge-socket-script-path
     'variable-documentation
     "ソケットのためのスクリプトのあるディレクトリを指定します。このディレク
トリが存在していなければ、自動的に作成されます。")

(put 'scim-bridge-socket-script-name
     'variable-documentation
     "スクリプトのファイル名を指定します。")

(put 'scim-bridge-socket-script
     'variable-documentation
     "スクリプトの内容です。このスクリプトは第1引数をソケット名として用いなけ
ればなりません。")

(put 'scim-meta-key-exists
     'variable-documentation
     "キーボードにmetaキーがある場合、この変数には t がセットされます。
自動検出がうまくいかない場合は、scim-bridge.el を読み込む前に手動で設定して
ください。")

(put 'scim-incompatible-mode-hooks
     'variable-documentation
     "これらのフックが実行される時には、scim-mode-map が無効化されます。")

(setcar (nthcdr 2 (symbol-function 'scim-define-common-key))
     "どのキーイベントが常時 SCIM へ送られるかを指定します。HANDLE が nil で
なければ、SCIM は KEY で与えられたキーイベントを処理します。KEY が配列として
与えられた場合は、キーシーケンスを表すのではなく、単独のキーストロークの複数
の定義を表します。
この設定を有効にするには、関数 `scim-update-key-bindings' を呼び出すか、
または scim-mode を再起動する必要があります。")

(setcar (nthcdr 2 (symbol-function 'scim-define-preedit-key))
     "どのキーイベントがプリエディット時に SCIM へ送られるかを指定します。
HANDLE が nil でなければ、SCIM は KEY で与えられたキーイベントを処理します。
KEY が配列として与えられた場合は、キーシーケンスを表すのではなく、単独のキー
ストロークの複数の定義を表します。
この設定を有効にするには、関数 `scim-update-key-bindings' を呼び出すか、
または scim-mode を再起動する必要があります。")

(provide 'scim-bridge-ja)

;;;
;;; scim-bridge-ja.el ends here
