;;; scim-bridge.el

;; Copyright (C) 2008 S. Irie

;; Author: S. Irie
;; Maintainer: S. Irie
;; Keywords: Input Method, i18n

(defconst scim-mode-version "0.6.1")

;; This program is free software; you can redistribute it and/or
;; modify it under the terms of the GNU General Public License as
;; published by the Free Software Foundation; either version 2, or
;; (at your option) any later version.

;; It is distributed in the hope that it will be useful, but WITHOUT
;; ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
;; or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public
;; License for more details.

;; You should have received a copy of the GNU General Public
;; License along with this program; if not, write to the Free
;; Software Foundation, Inc., 59 Temple Place, Suite 330, Boston,
;; MA 02111-1307 USA

;;; Commentary:

;; This program is scim-bridge client for GNU Emacs. It is, however,
;; not part of official scim-bridge.

;; The Smart Common Input Method platform (SCIM) is an input
;; method (IM) platform containing support for more than thirty
;; languages (CJK and many European languages) for POSIX-style
;; operating systems including Linux and BSD.

;;
;; Installation:
;;
;; First, save this file as scim-bridge.el and byte-compile in
;; a directory that is listed in load-path.
;;
;; Put the following in your .emacs file:
;;
;;   (require 'scim-bridge)
;;
;; After that, execute Emacs by typing on command line:
;;
;;   XMODIFIERS=@im=none emacs
;;
;; and turn on scim-mode:
;;
;;   M-x scim-mode
;;
;;
;; Here is the example of settings in .emacs:
;;
;;   (require 'scim-bridge)
;;   ;; Toggle the input mode of buffers individually
;;   (setq scim-mode-local t)
;;   ;; Use C-SPC for Set Mark command
;;   (scim-define-common-key ?\C-\  nil)
;;   ;; Use C-/ for Undo command
;;   (scim-define-common-key ?\C-/ nil)
;;   ;; Adjust horizontal position of candidate window
;;   (setq scim-adjust-window-x-position 'gnome)
;;   ;; Turn on scim-mode automatically
;;   (scim-mode t)
;;
;; Note that this program requires GNU Emacs 22 or later, and
;; doesn't work when Emacs is running on terminal emulator.
;;

;; History:
;; 2008-05-20  S. Irie
;;         * Bug fix
;;         * Version 0.6.1
;;
;; 2008-05-18  S. Irie
;;         * Manage meta-to-alt mapping
;;         * Bug fix
;;         * Version 0.6.0
;;
;; 2008-05-15  S. Irie
;;         * Experimental first release
;;         * Version 0.5.0
;;
;; 2008-05-12  S. Irie
;;         * Version 0.4.0
;;
;; 2008-05-08  S. Irie
;;         * Version 0.3.0
;;
;; 2008-05-06  S. Irie
;;         * Version 0.2.0
;;
;; 2008-04-20  S. Irie
;;         * Version 0.1.0

;; ToDo:

;;; Code:

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; User settings
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defgroup scim nil
  "The Smart Common Input Method platform"
  :prefix "scim-"
  :group 'editing :group 'wp)

;; Basic settings
(defgroup scim-basic nil
  "Settings of operation, such as mode management and keyboard"
  :group 'scim)

(defcustom scim-mode-local nil
  "If the value is non-nil, IMContexts are registered for each buffer
so that the input method of buffers can be toggled individually."
  :type 'boolean
  :group 'scim-basic)

(defun scim-customize-key (var value)
  (set var value)
  (if (fboundp 'scim-update-key-bindings)
      (scim-update-key-bindings var)))

(defcustom scim-common-function-key-list
  '((control ".")
    (control ",")
    (control "<")
    (control ">")
    (control "/")
    (control " ")
    (shift " ")
    (control alt left)
    (control alt right)
    (control alt up)
    (control alt down)
    (zenkaku-hankaku)
    (henkan)
    (shift henkan)
    (muhenkan)
    (hiragana-katakana)
    (alt romaji)
    (f6)
    (f7)
    (f8)
    (shift f8)
    (f9)
    (f10)
    (f11)
    (f12))
  "This list indicates which keystrokes SCIM takes over at both direct
insert mode and preediting mode. You can also add/remove the elements
using the function `scim-define-common-key'.
NOTICE: Don't set prefix keys in this option, such as ESC and C-x.
If you do so, operating Emacs might become impossible."
  :set 'scim-customize-key
  :type '(repeat (list :format "%v"
		       (set :format "%v"
			    :inline t
			    (const :format "M- " meta)
			    (const :format "C- " control)
			    (const :format "S- " shift)
			    (const :format "H- " hyper)
			    (const :format "s- " super)
			    (const :format "A- " alt))
		       (restricted-sexp :format "%v"
					:match-alternatives
					(symbolp stringp))))
  :group 'scim-basic)

(defcustom scim-preedit-function-key-list
  '((escape)
    (left)
    (right)
    (up)
    (down)
    (home)
    (end)
    (prior)
    (next)
    (return)
    (shift left)
    (shift right)
    (shift up)
    (shift down)
    (shift return)
    (tab)
    (iso-lefttab)
    (shift tab)
    (shift iso-lefttab)
    (backtab)
    (backspace)
    (delete)
    (kp-enter)
    (kp-space)
    (kp-tab)
    (kp-add)
    (kp-subtract))
  "This list indicates which keystrokes SCIM takes over when the
preediting area exists. You can also add/remove the elements using
the function `scim-define-preedit-key'."
  :set 'scim-customize-key
  :type '(repeat (list :format "%v"
		       (set :format "%v"
			    :inline t
			    (const :format "M- " meta)
			    (const :format "C- " control)
			    (const :format "S- " shift)
			    (const :format "H- " hyper)
			    (const :format "s- " super)
			    (const :format "A- " alt))
		       (restricted-sexp :format "%v"
					:match-alternatives
					(symbolp stringp))))
  :group 'scim-basic)

(defcustom scim-use-kana-ro-key nil
  "If you use Japanese kana typing method with jp-106 keyboard, turn
on this option to input a kana character `ろ' without pushing the
shift key.
 This option is made effectual by modifying the X-window system's
keyboard configurations using `xmodmap' command, so this affects all
other applications."
  :set 'scim-customize-key
  :type 'boolean
  :group 'scim-basic)

(defcustom scim-key-release-delay nil
  "If you use Japanese thumb shift typing method on SCIM-Anthy with
jp-106 keyboard, set the time delay of the key release (in seconds).
Set a smaller value than the simultaneous pressing time setting of
SCIM-Anthy."
  :type '(choice (const :tag "none" nil)
		 (number :tag "delay (in seconds)"
			 :value 0.1))
  :group 'scim-basic)

(defcustom scim-undo-by-committed-string nil
  "If the value is nil, undo is performed bringing some short
committed strings together or dividing the long committed string
within the range which does not exceed 20 characters. Othewse, undo
is executed every committed string."
  :type 'boolean
  :group 'scim-basic)

(defcustom scim-clear-preedit-when-unexpected-event nil
  "This option specifies the behavior for the cases that unexpected
event in preediting is caused. The unexpected event is, for example,
pasting the string with mouse. If the value is non-nil, the preediting
area is cleared."
  :type 'boolean
  :group 'scim-basic)

;; Appearance
(defgroup scim-appearance nil
  "Faces, candidate window, etc."
  :group 'scim)

(defface scim-preedit-default-face
;  nil
  '((t :inherit underline))
  "This face indicates the whole of the pre-editing area."
  :group 'scim-appearance)

(defface scim-preedit-underline-face
  '((t :inherit underline))
  ""
  :group 'scim-appearance)

(defface scim-preedit-highlight-face
;  '((t :inherit (scim-preedit-underline-face highlight)))
  '((t :inherit highlight))
  ""
  :group 'scim-appearance)

(defface scim-preedit-reverse-face
;  '((t :inherit scim-preedit-underline-face :inverse-video t))
  '((t :inverse-video t))
  ""
  :group 'scim-appearance)

(defcustom scim-adjust-window-x-position
;  'gnome
  nil
  "This option specifies whether the position of candidate window
is adjusted so that the inline candidate and the candidates in that
window may just line up in the vertical direction. If the value is
`gnome', the adjustment will be done using gnome's font size.
Otherwise, if the value is integer, that indicates the offset from
normal position by pixel number.
 This is not suitable for input method of the type to which the
candidate window is always displayed such as SCIM-pinyin (chinese),
because there is a possibility that the window hides the cursor when
the cursor is on the bottom of screen."
  :type '(choice (const :tag "use GNOME's font size" gnome)
		 (integer :tag "specify by pixel number" :value 24)
		 (const :tag "off" nil))
  :group 'scim-appearance)

(defcustom scim-adjust-window-y-position
  t
  "This option specifies whether the vertical position of candidate
window is adjusted to the cursor by using `xwininfo' command. If the
value is nil, the adjustment isn't done and therefore the position of
the window might slightly shift."
  :type 'boolean
  :group 'scim-appearance)

(defcustom scim-prediction-window-position
  '(nil . nil)
  "(Japanese IM only) The value should be given as (POS . ADJ).
If POS is non-nil, the forecast window is displayed under the head
of preedition area. If the value of ADJ is non-nil, x-position is
adjusted same as candidate window."
  :type '(cons
	       (choice :tag "Position"
		       (const :tag "Tail of preediting area" nil)
		       (const :tag "Head of preediting area" t))
	       (choice :tag "Adjustment"
		       (const :tag "same as conversion window" t)
		       (const :tag "off" nil)))
  :group 'scim-appearance)

(defcustom scim-mode-line-string " SCIM"
  "This variable controls enabling of the minor mode, and string is
a short string, starting with a space, to represent the mode in the
mode line. These strings must be short so that there is room for
several of them at once."
  :type 'string
  :group 'scim-appearance)

;; Advanced settings
(defgroup scim-expert nil
  "Advanced settings"
  :group 'scim)

(defcustom scim-focus-update-interval 0.3
  "The window focus is checked with this cycle measured in seconds.
It is necessary to restart scim-mode so that this setting may become
effective."
  :type 'number
  :group 'scim-expert)

(defcustom scim-bridge-timeout 500
  "Don't change this setting as long as there is no special reason."
  :type 'integer
  :group 'scim-expert)

(defcustom scim-bridge-wait-reply t
  "Don't change this setting as long as there is no special reason."
  :type 'boolean
  :group 'scim-expert)

(defcustom scim-bridge-socket-use-script nil
  "If the value is non-nil, the script is forced to use for connecting
to scim-bridge agent through UNIX domain socket instead of a built-in
function `make-network-process'. This is not usually required if your
Emacs' major version is 22 or later."
  :type 'boolean
  :group 'scim-expert)

(defcustom scim-bridge-socket-script-path "~/bin"
  "The destination of the socket script. When the directory doesn't
exist, it is made automatically."
  :type 'directory
  :group 'scim-expert)

(defcustom scim-bridge-socket-script-name "scim-bridge-client.pl"
  "The filename of supplementary script."
  :type 'string
  :group 'scim-expert)

(defcustom scim-bridge-socket-script
  "#!/usr/bin/perl

use Socket;
use strict;
use FileHandle;
use threads;

my $name = shift;

socket(SOCKET, PF_UNIX, SOCK_STREAM, 0) || die \"socket: $!\";
connect(SOCKET, sockaddr_un($name)) || die \"connect: $!\";

my $thread = threads->new(\\&receive);

SOCKET->autoflush(1);
while (<>) {
    print SOCKET $_;
}
$thread->join;
exit;

sub receive {
    while (<SOCKET>) {
	print $_;
    }
}"
  "This is the real content of the supplementary script. The first
command line argument must be used as the socket name."
  :type 'string
  :group 'scim-expert)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; System settings and constants
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar scim-debug nil)
(defvar scim-log-buffer "*scim-bridge-log*")

;(defvar scim-bridge-socket-path "/tmp/scim-bridge-0.3.0.socket-1000@localhost:0.0")
(defvar scim-bridge-compat-version "0.3.0")
(defvar scim-bridge-socket-dir "/tmp/")
(defvar scim-bridge-socket-name "socket")
(defvar scim-bridge-name "scim-bridge")
(defvar scim-bridge-host-name "localhost")
(defvar scim-bridge-socket-path
  (let ((disp (getenv "DISPLAY")))
    (concat scim-bridge-socket-dir scim-bridge-name "-"
	    scim-bridge-compat-version "." scim-bridge-socket-name "-"
	    (number-to-string (user-uid)) "@" scim-bridge-host-name
	    (substring disp (string-match ":" disp)))))

(defvar scim-meta-key-exists
  (string< "" (shell-command-to-string "xmodmap -pke | grep '= Meta'"))
  "t is set in this variable if there is mata modifier key in the
keyboard. When automatic detection doesn't go well, please set the
value manually before scim-bridge.el is loaded.")

(defvar scim-tmp-buffer-name " *scim-bridge*"
  "This is working buffer name used for communicating with the agent.")

(defvar scim-incompatible-mode-hooks
  '(ediff-mode-hook)
  "When these hooks run, scim-mode-map become invalid.")

(defvar scim-reply-alist
  '(
    ;; Status
    ("imengine_status_changed"		. scim-imengine-status-changed)
    ("imcontext_registered"		. scim-imcontext-registered)
    ("preedit_mode_changed"		. scim-preedit-mode-changed)
    ("focus_changed"			. scim-focus-changed)
    ("cursor_location_changed"		. scim-cursor-location-changed)
    ("key_event_handled"		. scim-key-event-handled)
    ("imcontext_deregister"		. scim-imcontext-deregister)
    ("imcontext_reseted"		. scim-imcontext-reseted)
    ;; Request
    ("forward_key_event"		. scim-forward-key-event)
    ("update_preedit"			. scim-update-preedit)
    ("set_preedit_shown"		. scim-set-preedit-shown)
    ("set_preedit_cursor_position"	. scim-set-preedit-cursor-position)
    ("set_preedit_string"		. scim-set-preedit-string)
    ("set_preedit_attributes"		. scim-set-preedit-attributes)
    ("set_commit_string"		. scim-set-commit-string)
    ("commit_string"			. scim-commit-string)
    ("get_surrounding_text"		. scim-get-surrounding-text)
    ("delete_surrounding_text"		. scim-delete-surrounding-text)
    ("replace_surrounding_text"		. scim-replace-surrounding-text)
    ("beep"				. scim-beep)
    ))

(defvar scim-modifier-alist
  `(
    ;; Keyboard
    (shift . "shift")
    (control . "control")
    (,(if scim-meta-key-exists 'alt 'meta) . "alt")
    (meta . "meta")
;    (super . "hyper")
    (super . "super")
    (hyper . "hyper")
    ;;
    (caps-lock . "caps_lock")
    (num-lock . "num_lock")
    (kana-RO . "kana_ro")
    ;; Mouse
;    (up . "up")
;    (down . "down")
;    (drag . "drag")
;    (click . "click")
;    (double . "double")
;    (triple . "triple")
    ))

(defvar scim-alt-modifier-alist
  '(
    (hiragana-katakana . romaji)
    (zenkaku-hankaku . kanji)
    (henkan . mode-switch)
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Key code table
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defvar scim-keycode-alist
  '(
    ;; *** Function keys ***********************************************
    (backspace . ?\xff08)
    (tab . ?\xff09)
    (linefeed . ?\xff0a)
    (clear . ?\xff0b)
    (return . ?\xff0d)
    (pause . ?\xff13)
;    (scroll-lock . ?\xff14)
;    (sys-req . ?\xff15)
    (escape . ?\xff1b)
    (delete . ?\xffff)
    ;; *** International & multi-key character composition *************
;    (multi-key . ?\xff20)
;    (codeinput . ?\xff37)
;    (singlecandidate . ?\xff3c)
;    (multiplecandidate . ?\xff3d)
;    (previouscandidate . ?\xff3e)
    ; Japanese keyboard support
    (kanji . ?\xff21)
    (muhenkan . ?\xff22)
;    (henkan-mode . ?\xff23)
    (henkan . ?\xff23)
    (romaji . ?\xff24)
    (hiragana . ?\xff25)
    (katakana . ?\xff26)
    (hiragana-katakana . ?\xff27)
    (zenkaku . ?\xff28)
    (hankaku . ?\xff29)
    (zenkaku-hankaku . ?\xff2a)
    (touroku . ?\xff2b)
    (massyo . ?\xff2c)
    (kana-lock . ?\xff2d)
    (kana-shift . ?\xff2e)
    (eisu-shift . ?\xff2f)
    (eisu-toggle . ?\xff30)
;    (kanji-bangou . ?\xff37)
;    (zen-koho . ?\xff3d)
;    (mae-koho . ?\xff3e)
    ;; *** Cursor control & motion *************************************
    (home . ?\xff50)
    (left . ?\xff51)
    (up . ?\xff52)
    (right . ?\xff53)
    (down . ?\xff54)
    (prior . ?\xff55)
;    (page-up . ?\xff55)
    (next . ?\xff56)
;    (page-down . ?\xff56)
    (end . ?\xff57)
    (begin . ?\xff58)
    ; Misc Functions
    (select . ?\xff60)
    (print . ?\xff61)
    (execute . ?\xff62)
    (insert . ?\xff63)
    (undo . ?\xff65)
    (redo . ?\xff66)
    (menu . ?\xff67)
    (find . ?\xff68)
    (cancel . ?\xff69)
    (help . ?\xff6a)
    (break . ?\xff6b)
    (mode-switch . ?\xff7e) ;; This key cannot be recognized to Emacs.
;    (num-lock . ?\xff7f)
    ;; *** Keypad ******************************************************
    (kp-space . ?\xff80)
    (kp-tab . ?\xff89)
    (kp-enter . ?\xff8d)
    (kp-f1 . ?\xff91)
    (kp-f2 . ?\xff92)
    (kp-f3 . ?\xff93)
    (kp-f4 . ?\xff94)
    (kp-home . ?\xff95)
    (kp-left . ?\xff96)
    (kp-up . ?\xff97)
    (kp-right . ?\xff98)
    (kp-down . ?\xff99)
    (kp-prior . ?\xff9a)
;    (kp-page-up . ?\xff9a)
    (kp-next . ?\xff9b)
;    (kp-page-down . ?\xff9b)
    (kp-end . ?\xff9c)
    (kp-begin . ?\xff9d)
    (kp-insert . ?\xff9e)
    (kp-delete . ?\xff9f)
    (kp-equal . ?\xffbd)
    (kp-multiply . ?\xffaa)
    (kp-add . ?\xffab)
    (kp-separator . ?\xffac)
    (kp-subtract . ?\xffad)
    (kp-decimal . ?\xffae)
    (kp-divide . ?\xffaf)
    (kp-0 . ?\xffb0)
    (kp-1 . ?\xffb1)
    (kp-2 . ?\xffb2)
    (kp-3 . ?\xffb3)
    (kp-4 . ?\xffb4)
    (kp-5 . ?\xffb5)
    (kp-6 . ?\xffb6)
    (kp-7 . ?\xffb7)
    (kp-8 . ?\xffb8)
    (kp-9 . ?\xffb9)
    ;; *** Auxilliary functions ****************************************
    (f1 . ?\xffbe)
    (f2 . ?\xffbf)
    (f3 . ?\xffc0)
    (f4 . ?\xffc1)
    (f5 . ?\xffc2)
    (f6 . ?\xffc3)
    (f7 . ?\xffc4)
    (f8 . ?\xffc5)
    (f9 . ?\xffc6)
    (f10 . ?\xffc7)
    (f11 . ?\xffc8)
    (f12 . ?\xffc9)
    (f13 . ?\xffca)
    (f14 . ?\xffcb)
    (f15 . ?\xffcc)
    (f16 . ?\xffcd)
    (f17 . ?\xffce)
    (f18 . ?\xffcf)
    (f19 . ?\xffd0)
    (f20 . ?\xffd1)
    (f21 . ?\xffd2)
    (f22 . ?\xffd3)
    (f23 . ?\xffd4)
    (f24 . ?\xffd5)
    (f25 . ?\xffd6)
    (f26 . ?\xffd7)
    (f27 . ?\xffd8)
    (f28 . ?\xffd9)
    (f29 . ?\xffda)
    (f30 . ?\xffdb)
    (f31 . ?\xffdc)
    (f32 . ?\xffdd)
    (f33 . ?\xffde)
    (f34 . ?\xffdf)
    (f35 . ?\xffe0)
    ;; *** Modifier keys ***********************************************
;    (shift-l . ?\xffe1)
;    (shift-r . ?\xffe2)
;    (control-l . ?\xffe3)
;    (control-r . ?\xffe4)
;    (caps-lock . ?\xffe5)
    (capslock . ?\xffe5)
;    (shift-lock . ?\xffe6)
;    (meta-l . ?\xffe7)
;    (meta-r . ?\xffe8)
;    (alt-l . ?\xffe9)
;    (alt-r . ?\xffea)
;    (super-l . ?\xffeb)
;    (super-r . ?\xffec)
;    (hyper-l . ?\xffed)
;    (hyper-r . ?\xffee)
    ;; *** ISO 9995 function and modifier keys *************************
;    (iso-lock . ?\xfe01)
;    (iso-level2-latch . ?\xfe02)
;    (iso-level3-shift . ?\xfe03)
;    (iso-level3-latch . ?\xfe04)
;    (iso-level3-lock . ?\xfe05)
;    (iso-group-shift . ?\xff7e)
;    (iso-group-latch . ?\xfe06)
;    (iso-group-lock . ?\xfe07)
;    (iso-next-group . ?\xfe08)
;    (iso-next-group-lock . ?\xfe09)
;    (iso-prev-group . ?\xfe0a)
;    (iso-prev-group-lock . ?\xfe0b)
;    (iso-first-group . ?\xfe0c)
;    (iso-first-group-lock . ?\xfe0d)
;    (iso-last-group . ?\xfe0e)
;    (iso-last-group-lock . ?\xfe0f)
;    (iso-left-tab . ?\xfe20)
    (iso-lefttab . ?\xfe20)
    (iso-move-line-up . ?\xfe21)
    (iso-move-line-down . ?\xfe22)
    (iso-partial-line-up . ?\xfe23)
    (iso-partial-line-down . ?\xfe24)
    (iso-partial-space-left . ?\xfe25)
    (iso-partial-space-right . ?\xfe26)
    (iso-set-margin-left . ?\xfe27)
    (iso-set-margin-right . ?\xfe28)
    (iso-release-margin-left . ?\xfe29)
    (iso-release-margin-right . ?\xfe2a)
    (iso-release-both-margins . ?\xfe2b)
    (iso-fast-cursor-left . ?\xfe2c)
    (iso-fast-cursor-right . ?\xfe2d)
    (iso-fast-cursor-up . ?\xfe2e)
    (iso-fast-cursor-down . ?\xfe2f)
    (iso-continuous-underline . ?\xfe30)
    (iso-discontinuous-underline . ?\xfe31)
    (iso-emphasize . ?\xfe32)
    (iso-center-object . ?\xfe33)
    (iso-enter . ?\xfe34)
    ;; *** Lispy accent keys *******************************************
    (dead-grave . ?\xfe50)
    (dead-acute . ?\xfe51)
    (dead-circumflex . ?\xfe52)
    (dead-tilde . ?\xfe53)
    (dead-macron . ?\xfe54)
    (dead-breve . ?\xfe55)
    (dead-abovedot . ?\xfe56)
    (dead-diaeresis . ?\xfe57)
    (dead-abovering . ?\xfe58)
    (dead-doubleacute . ?\xfe59)
    (dead-caron . ?\xfe5a)
    (dead-cedilla . ?\xfe5b)
    (dead-ogonek . ?\xfe5c)
    (dead-iota . ?\xfe5d)
    (dead-voiced-sound . ?\xfe5e)
    (dead-semivoiced-sound . ?\xfe5f)
    (dead-belowdot . ?\xfe60)
    (dead-hook . ?\xfe61)
    (dead-horn . ?\xfe62)
    ;; *** Katakana ****************************************************
    (overline . ?\x47e)
    (kana-fullstop . ?\x4a1)
    (kana-openingbracket . ?\x4a2)
    (kana-closingbracket . ?\x4a3)
    (kana-comma . ?\x4a4)
    (kana-conjunctive . ?\x4a5)
;    (kana-middledot . ?\x4a5)
    (kana-WO . ?\x4a6)
    (kana-a . ?\x4a7)
    (kana-i . ?\x4a8)
    (kana-u . ?\x4a9)
    (kana-e . ?\x4aa)
    (kana-o . ?\x4ab)
    (kana-ya . ?\x4ac)
    (kana-yu . ?\x4ad)
    (kana-yo . ?\x4ae)
    (kana-tsu . ?\x4af)
;    (kana-tu . ?\x4af)
    (prolongedsound . ?\x4b0)
    (kana-A . ?\x4b1)
    (kana-I . ?\x4b2)
    (kana-U . ?\x4b3)
    (kana-E . ?\x4b4)
    (kana-O . ?\x4b5)
    (kana-KA . ?\x4b6)
    (kana-KI . ?\x4b7)
    (kana-KU . ?\x4b8)
    (kana-KE . ?\x4b9)
    (kana-KO . ?\x4ba)
    (kana-SA . ?\x4bb)
    (kana-SHI . ?\x4bc)
    (kana-SU . ?\x4bd)
    (kana-SE . ?\x4be)
    (kana-SO . ?\x4bf)
    (kana-TA . ?\x4c0)
    (kana-CHI . ?\x4c1)
;    (kana-TI . ?\x4c1)
    (kana-TSU . ?\x4c2)
;    (kana-TU . ?\x4c2)
    (kana-TE . ?\x4c3)
    (kana-TO . ?\x4c4)
    (kana-NA . ?\x4c5)
    (kana-NI . ?\x4c6)
    (kana-NU . ?\x4c7)
    (kana-NE . ?\x4c8)
    (kana-NO . ?\x4c9)
    (kana-HA . ?\x4ca)
    (kana-HI . ?\x4cb)
    (kana-FU . ?\x4cc)
;    (kana-HU . ?\x4cc)
    (kana-HE . ?\x4cd)
    (kana-HO . ?\x4ce)
    (kana-MA . ?\x4cf)
    (kana-MI . ?\x4d0)
    (kana-MU . ?\x4d1)
    (kana-ME . ?\x4d2)
    (kana-MO . ?\x4d3)
    (kana-YA . ?\x4d4)
    (kana-YU . ?\x4d5)
    (kana-YO . ?\x4d6)
    (kana-RA . ?\x4d7)
    (kana-RI . ?\x4d8)
    (kana-RU . ?\x4d9)
    (kana-RE . ?\x4da)
    (kana-RO . ?\x4db)
    (kana-WA . ?\x4dc)
    (kana-N . ?\x4dd)
    (voicedsound . ?\x4de)
    (semivoicedsound . ?\x4df)
;    (kana-switch . ?\xFF7E)
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Definition of variables
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Mode management
(defcustom scim-mode nil
  "Toggle scim-mode.
Setting this variable directly does not take effect;
use either \\[customize] or the function `scim-mode'."
  :set 'custom-set-minor-mode
  :initialize 'custom-initialize-default
  :version "22.1"
  :type 'boolean
  :group 'scim
  :require 'scim-bridge)

(defvar scim-mode-map nil)
(defvar scim-mode-preedit-map nil)
(defvar scim-mode-map-alist nil)
(defvar scim-mode-map-disabled nil)
(make-variable-buffer-local 'scim-mode-map-disabled)

(defvar scim-bridge-socket nil)
(defvar scim-tmp-buffer nil)
(defvar scim-last-command-event nil)
(defvar scim-last-command-buffer nil)
(defvar scim-buffer-undo-list nil)
(defvar scim-buffer-modified-p nil)
(defvar scim-frame-focus nil)
(defvar scim-focus-update-timer nil)

;; Preedit area
(defvar scim-imcontext-id nil)
(defvar scim-preedit-buffer nil)
(defvar scim-preedit-point 0)
(defvar scim-preedit-update nil)
(defvar scim-preedit-shown "")
(defvar scim-preedit-string "")
(defvar scim-preedit-prev-string "")
(defvar scim-preedit-curpos 0)
(defvar scim-preedit-attributes nil)
(defvar scim-preedit-default-attr nil)
(defvar scim-preedit-overlays nil)
(defvar scim-committed-string "")
(defvar scim-frame-extents nil)
(defvar scim-adjust-window-x-offset 0)
(defvar scim-adjust-window-y-offset 0)
(defvar scim-surrounding-text-modified nil)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Definition of functions
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Miscellaneous
(defun scim-escape-string (str)
  (let* ((tmp (append str nil))
	 cur
	 (next tmp))
    (while (setq cur (memq ?\\ next))
      (setq next (cdr cur))
      (setcdr cur (cons ?\\ next)))
    (setq next tmp)
    (while (setq cur (memq ?\n next))
      (setq next (cdr cur))
      (setcar cur ?\\)
      (setcdr cur (cons ?n next)))
    (setq next tmp)
    (while (setq cur (memq ?\  next))
      (setq next (cdr cur))
      (setcar cur ?\\)
      (setcdr cur (cons ?s next)))
    (concat tmp)))

(defun scim-construct-command (list)
  (mapconcat (lambda (f) (scim-escape-string f)) list " "))

(defun scim-unescape-string (str)
  (let* ((tmp (append str nil))
	 (cur tmp)
	 next)
    (while (setq cur (memq ?\\ cur))
      (setq next (cdr cur))
      (setcar cur (car (or (rassq (car next)
				  '((?\  . ?s)
				    (?\n . ?n)
;				    (?\\ . ?\\)
				    ))
			   next)))
      (setq cur (setcdr cur (cdr next))))
    (concat (delq nil tmp))))

(defun scim-split-commands (commands)
  (mapcar (lambda (line)
	    (mapcar 'scim-unescape-string (split-string line " ")))
	  (split-string (substring commands 0 -1) "\n")))

(defun scim-decode-event (event)
  ;; Convert Emacs event to scim-bridge command
  (let ((modifiers (mapcar (lambda (mod)
			     (cdr (assq mod scim-modifier-alist)))
			   (event-modifiers event)))
	(key-code (event-basic-type event)))
    (if (numberp key-code)
	(if (and (member "shift" modifiers)
		 (>= key-code ?a)
		 (<= key-code ?z))
	    (setq key-code (- key-code 32)))
      (if (member "alt" modifiers)
	  (setq key-code
		(or (cdr (assq key-code scim-alt-modifier-alist))
		    key-code)))
      (setq key-code (or (cdr (assq key-code scim-keycode-alist))
			 key-code)))
    (cons key-code modifiers)))

(defun scim-encode-event (key-code modifiers)
  ;; Convert scim-bridge command to Emacs event
  (setq key-code (string-to-number key-code))
  (let* ((bas (or (car (rassq key-code scim-keycode-alist))
		  (if (< key-code 128) key-code)))
	 (mods nil))
    (if (member "alt" modifiers)
	(setq bas (or (car (rassq bas scim-alt-modifier-alist))
		      bas)))
    (while modifiers
      (let ((m (car (rassoc (car modifiers) scim-modifier-alist))))
	(if m (cond ((eq m 'caps-lock)
		     nil) ;; Ignore `caps_lock' modifier
		    ((eq m 'num-lock)
		     nil) ;; Ignore `num_lock' modifier
		    ((eq m 'kana-RO)
		     (if (and scim-use-kana-ro-key (eq bas ?\\))
			 (setq bas ?_)))
		    (t (add-to-list 'mods m)))))
      (setq modifiers (cdr modifiers)))
    (if bas (event-convert-list (nconc mods (list bas))))))

(defconst scim-hex-table
  (vconcat (make-vector ?0 0)
	   [0 1 2 3 4 5 6 7 8 9]
	   (make-vector (- ?A ?9 1) 0)
	   [10 11 12 13 14 15]
	   (make-vector (- ?a ?F 1) 0)
	   [10 11 12 13 14 15]
	   (make-vector (- 127 ?f) 0)))

(defun scim-hexstr-to-number (string)
  (let ((len (length string))
	(ret 0)	(pos 0))
    (while (< (setq ret (+ (* ret 16)
			   (aref scim-hex-table (aref string pos)))
		    pos (1+ pos))
	      len))
    ret))

;(defun scim-hexstr-to-number (string &optional length)
;  (let ((pos (1- (or length (length string)))))
;    (+ (if (= pos 0)
;	   0
;	 (* 16 (scim-hexstr-to-number string pos)))
;       (aref scim-hex-table (aref string pos)))))

(defun scim-twos-complement (string)
  (let ((num (string-to-number string)))
    (if (< num 2147483648.0)
	(round num)
      (round (- num 4294967296.0)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Message & Log
(defmacro scim-message (format-string &rest args)
  `(if (not scim-debug)
       (message ,format-string ,@args)
     (let ((log-str (format ,format-string ,@args)))
       (save-excursion
	 (set-buffer (get-buffer-create scim-log-buffer))
	 (goto-char (point-max))
	 (insert (concat (format log-str) "\n"))))))

(defmacro scim-show-undo-list (format-string &rest args)
  `(progn
     (scim-message ,format-string ,@args)
     (if (not (listp buffer-undo-list))
	 (scim-message "undo list (disabled): %s" buffer-undo-list)
       (scim-message " top: %s" (car buffer-undo-list))
       (scim-message " 2nd: %s" (car (cdr buffer-undo-list)))
       (scim-message " 3rd: %s" (car (cdr (cdr buffer-undo-list))))
       (scim-message " 4th: %s" (car (cdr (cdr (cdr buffer-undo-list))))))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Control buffer-undo-list
(defun scim-disable-undo ()
  (unless scim-preedit-buffer
    (setq scim-buffer-undo-list buffer-undo-list
	  scim-buffer-modified-p (buffer-modified-p))
    (buffer-disable-undo)
    (setq scim-preedit-buffer (current-buffer))
    (if scim-debug (scim-message "disable undo list: %s" scim-preedit-buffer))
    ))

(defun scim-enable-undo ()
  (when scim-preedit-buffer
    (with-current-buffer scim-preedit-buffer
      (buffer-enable-undo)
      (setq buffer-undo-list scim-buffer-undo-list)
      (set-buffer-modified-p scim-buffer-modified-p))
    (if scim-debug (scim-message "enable undo list: %s" scim-preedit-buffer))
    (setq scim-preedit-buffer nil)))

(defun scim-insert-and-modify-undo-list (str)
  (let* ((prev-list (if (car-safe buffer-undo-list)
			buffer-undo-list
		      (cdr-safe buffer-undo-list)))
	 (prev (car-safe prev-list))
	 (consp-prev (and (consp prev)
			  (integerp (car prev))
			  (integerp (cdr prev))))
	 (consecutivep (and consp-prev
			   (= (cdr prev) (point))
			   (not (= (preceding-char) ?\n))
			   (<= (+ (cdr prev) (length str))
			       (+ (car prev) 20))))) ;; max 20 chars
    (if scim-debug (scim-show-undo-list "previous undo list"))
    (when (and consp-prev
	       (integerp (car (cdr prev-list))))
      (setcdr prev-list (cdr (cdr prev-list)))
      (if scim-debug (scim-show-undo-list "get rid of point setting entry"))
      )
    (insert str)
    (if scim-debug (scim-show-undo-list "insert string: %S" str))
    (when (integerp (car (cdr-safe buffer-undo-list)))
      (setcdr buffer-undo-list (cdr (cdr buffer-undo-list)))
      (if scim-debug (scim-show-undo-list "get rid of point setting entry"))
      )
    (if (and consecutivep
	     (eq (cdr (cdr buffer-undo-list)) prev-list))
	(progn
	  (setcar (car buffer-undo-list) (car (car prev-list)))
	  (setcdr buffer-undo-list (cdr prev-list))
	  (if scim-debug (scim-show-undo-list "unify consecutive insertion entries"))
	  )
      (when (> (length str) 20)
	(let ((beg (car (car buffer-undo-list)))
	      (end (cdr (car buffer-undo-list)))
	      (new-list (cdr buffer-undo-list)))
	  (while (> (- end beg) 20)
	    (setq new-list (cons nil (cons (cons beg (+ beg 20)) new-list))
		  beg (+ beg 20)))
	  (setq buffer-undo-list (cons (cons beg end) new-list))
	  (if scim-debug (scim-show-undo-list "divide long insertion entry"))
	  )))
    ))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Control keyboard
(defun scim-enable-keymap ()
  (setq minor-mode-overriding-map-alist
	(delete '(scim-mode) minor-mode-overriding-map-alist))
  (setq scim-mode-map-disabled nil))

(defun scim-disable-keymap ()
  (add-to-list 'minor-mode-overriding-map-alist '(scim-mode))
  (setq scim-mode-map-disabled t))

(defun scim-enable-preedit-keymap ()
  (setq minor-mode-map-alist
	(cons (cons 'scim-mode scim-mode-preedit-map)
	      (delq (assq 'scim-mode minor-mode-map-alist)
		    minor-mode-map-alist))
	scim-mode-map-alist
	(list (assoc 'scim-mode minor-mode-map-alist))))

(defun scim-disable-preedit-keymap ()
  (setcdr (assq 'scim-mode minor-mode-map-alist)
	  scim-mode-map)
  (setq scim-mode-map-alist nil))

(defun scim-enable-kana-ro-key ()
  (if scim-debug (scim-message "enable Kana-RO key"))
  (shell-command-to-string
   "xmodmap -pke | sed -n 's/= backslash\\( underscore\\)/=\\1\\1/p' | xmodmap -"))

(defun scim-disable-kana-ro-key ()
  (if scim-debug (scim-message "disable Kana-RO key"))
  (shell-command-to-string
   "xmodmap -pke | sed -n 's/=\\( underscore\\)\\1/= backslash\\1/p' | xmodmap -"))

(defun scim-update-kana-ro-key ()
  (if (and window-system
	   (not (string=
		 (shell-command-to-string
		  "xprop -root | grep '^_XKB_RULES_NAMES.*\"jp106\"'")
		 "")))
      (if scim-use-kana-ro-key
	  (scim-enable-kana-ro-key)
	(scim-disable-kana-ro-key))))

(defun scim-make-keymap-internal (map min max keys)
  (let ((i min))
    (while (<= i max)
      (define-key map (char-to-string i) 'scim-handle-event)
      (setq i (1+ i))))
  (while keys
    (let* ((key (copy-sequence (car keys)))
	   (n (1- (length key)))
	   (bas (nth n key)))
      (if (stringp bas)
	  (setcar (nthcdr n key) (string-to-char bas)))
      (when (memq 'alt key)
	(unless scim-meta-key-exists
	  (setq key (cons 'meta (delq 'alt key))))
	(setcar (nthcdr n key)
		(or (car (rassq bas scim-alt-modifier-alist))
		    bas)))
      (define-key map (vector key) 'scim-handle-event)
      (setq keys (cdr keys))))
  map)

(defun scim-make-common-map ()
  (scim-make-keymap-internal (make-keymap)
			     32 126
			     scim-common-function-key-list))

(defun scim-make-preedit-map ()
  (scim-make-keymap-internal (scim-make-common-map)
			     0 31
			     scim-preedit-function-key-list))

(defun scim-update-key-bindings (&optional symbol)
  (if (or (not symbol)
	  (eq symbol 'scim-use-kana-ro-key))
      (scim-update-kana-ro-key))
  (when (or (not symbol)
	    (eq symbol 'scim-common-function-key-list)
	    (eq symbol 'scim-preedit-function-key-list))
    (unless (eq symbol 'scim-preedit-function-key-list)
      (if scim-debug (scim-message "update scim-mode-map"))
      (setq scim-mode-map (scim-make-common-map)))
    (if scim-debug (scim-message "update scim-mode-preedit-map"))
    (setq scim-mode-preedit-map (scim-make-preedit-map))
;    (if scim-debug (scim-message "scim-mode-preedit-map: %s" scim-mode-preedit-map))
    (if scim-preedit-buffer
	(scim-enable-preedit-keymap)
      (scim-disable-preedit-keymap))))

(defun scim-define-key (symbol keys handle)
  ;; If keys is given as an array, it doesn't indicate key sequence,
  ;; but multiple definitions of single keystroke.
  (let ((keys-list (if (arrayp keys)
		       (listify-key-sequence keys)
		     (list keys))))
    (while keys-list
      (let ((key (car keys-list)))
	(if (listp key)
	    (let* ((n (1- (length key)))
		   (bas (nth n key)))
	      ;; If the key event is specified by a list and the last
	      ;; element is given as a string, the code number for the first
	      ;; character of the string is used for an event basic type.
	      (when (stringp bas)
		(setq key (copy-sequence key))
		(setcar (nthcdr n key) (string-to-char bas)))
	      (setq key (event-convert-list key))))
	;; It is necessary to call a function event-basic-type after
	;; event-modifiers because event-basic-type uses the symbol
	;; property event-symbol-elements added by event-modifiers
	;; when event is given as a symbol.
	(let ((modifiers (event-modifiers key))
	      (key-code (event-basic-type key)))
	  (if (integerp key-code)
	      (setq key-code (char-to-string key-code)
		    modifiers (reverse modifiers)))
	  (setq key (append modifiers (list key-code))))
	(if handle
	    (add-to-list symbol key)
	  (set symbol (delete key (symbol-value symbol)))))
      (setq keys-list (cdr keys-list)))
    (symbol-value symbol))) ;; Return value

(defun scim-define-common-key (key handle)
  "Specify which key events SCIM anytime takes over. If HANDLE
is non-nil, SCIM handles the key events given by KEY. When KEY is
given as an array, it doesn't indicate key sequence, but multiple
definitions of single keystroke.
 It is necessary to call a function `scim-update-key-bindings' or
restart scim-mode so that this settings may become effective."
  (scim-define-key 'scim-common-function-key-list key handle))

(defun scim-define-preedit-key (key handle)
  "Specify which key events SCIM takes over when preediting. If
HANDLE is non-nil, SCIM handles the key events given by KEY. When
KEY is given as an array, it doesn't indicate key sequence, but
multiple definitions of single keystroke.
 It is necessary to call a function `scim-update-key-bindings' or
restart scim-mode so that this settings may become effective."
  (scim-define-key 'scim-preedit-function-key-list key handle))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Control display
(defun scim-update-mode-line ()
  (force-mode-line-update)
  scim-mode) ;; Return value

;(defun scim-title-bar-height ()
;  "Return the pixel hight of title bar of selected frame."
;  (let* ((window-id (frame-parameter nil 'outer-window-id))
;	 (line (shell-command-to-string
;		(concat "xprop -id " window-id " _NET_FRAME_EXTENTS"))))
;    (string-to-number (substring line
;				 (string-match "[0-9]+,[ 0-9]+$" line)
;				 (string-match ",[ 0-9]+$" line)))))

(defun scim-get-frame-extents ()
  "Return the pixel width of frame edges as (left right top bottom).
Here, `top' also indicates the hight of title bar."
  (if scim-debug (scim-message "get frame extents"))
  (let* ((window-id (frame-parameter nil 'outer-window-id))
	 (line (shell-command-to-string
		(concat "xprop -id " window-id " _NET_FRAME_EXTENTS"))))
    (mapcar 'string-to-number
	    (split-string (substring line (string-match "[0-9]" line) -1)
			  ","))))

(defun scim-save-frame-extents ()
  (setq scim-frame-extents (scim-get-frame-extents)))

(defun scim-frame-header-height ()
  "Return the total of pixel height of menu-bar and tool-bar."
  (- (frame-pixel-height)
     (* (frame-height) (frame-char-height))
     scim-adjust-window-y-offset))

(defun scim-real-frame-header-height ()
  (if scim-debug (scim-message "get frame header height"))
  (let* ((window-id (frame-parameter nil 'window-id))
	 (line (shell-command-to-string
		(concat "xwininfo -id " window-id
			" | grep 'Relative upper-left Y:'"))))
    (string-to-number (substring line
				 (string-match "[0-9]+$" line)))))

(defun scim-set-window-y-offset ()
  (setq scim-adjust-window-y-offset
	(if scim-adjust-window-y-position
	    (let ((scim-adjust-window-y-offset 0))
	      (- (scim-frame-header-height)
		 (scim-real-frame-header-height)))
	  0)))

(defun scim-compute-pixel-position ()
  "Return the screen pxel position of point as (X . Y).
Its values show the coordinates of lower left corner of the character."
;  (if scim-debug (scim-message "current-buffer: %s  pos: %d" (current-buffer) (point)))
  (let* ((x-y (or (posn-x-y (posn-at-point))
		  '(0 . 0))))
;    (if scim-debug (scim-message "(x . y): %s" x-y))
    (cons (+ (frame-parameter nil 'left)
	     (car scim-frame-extents)
	     (car (window-inside-pixel-edges))
	     (car x-y))
	  (+ (frame-parameter nil 'top)
	     (nth 2 scim-frame-extents)
	     (scim-frame-header-height)
	     (car (cdr (window-pixel-edges)))
	     (cdr x-y)
	     (frame-char-height)))))

(defun scim-get-gnome-font-size ()
  (if (string= (shell-command-to-string "which gconftool-2") "")
      0
    (let ((font (shell-command-to-string
		 "gconftool-2 -g /desktop/gnome/interface/font_name"))
	  (dpi (shell-command-to-string
		"gconftool-2 -g /desktop/gnome/font_rendering/dpi")))
      (/ (* (string-to-number
	     (substring font (string-match "[0-9]+$" font) -1))
	    (string-to-number dpi))
	 72))))

(defun scim-set-window-x-offset ()
  (setq scim-adjust-window-x-offset
	(cond ((eq scim-adjust-window-x-position 'gnome)
	       (+ (scim-get-gnome-font-size) 4))
	      ((integerp scim-adjust-window-x-position)
	       scim-adjust-window-x-position)
	      (t 0))))

(defun scim-get-active-window-id ()
  "Return the number of the window-system window which is foreground,
i.e. input focus is in this window."
  (let ((line (shell-command-to-string "xprop -root _NET_ACTIVE_WINDOW")))
    (scim-hexstr-to-number
     (substring line (string-match "[0-9a-fA-F]+$" line) -1))))

(defun scim-check-frame-focus ()
  (if (stringp scim-imcontext-id)
      (let ((window-id (string-to-number
			(frame-parameter nil 'outer-window-id)))
	    (active-win (scim-get-active-window-id)))
	(when (and (eq (eq window-id active-win) (not scim-frame-focus))
		   (not isearch-mode))
	  (setq scim-frame-focus (not scim-frame-focus))
	  (if scim-debug (scim-message "change focus"))
	  (scim-change-focus scim-frame-focus)
	  (scim-save-frame-extents)
	  (scim-set-window-x-offset)))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Manipulate preedit buffer
(defun scim-check-rgb-color (string)
  (and (eq (length string) 7)
       (string-match "^#[0-9A-Fa-f]+$" string)))

(defun scim-select-face-by-attr (attr)
  (cdr (assoc attr
	      '(("underline" . scim-preedit-underline-face)
		("highlight" . scim-preedit-highlight-face)
		("reverse" . scim-preedit-reverse-face)))))

(defun scim-remove-preedit ()
  (remove-hook 'before-change-functions 'scim-before-change-function)
  (unless (string= scim-preedit-prev-string "")
    (delete-region scim-preedit-point
		   (+ scim-preedit-point (length scim-preedit-prev-string))))
  (mapc 'delete-overlay scim-preedit-overlays)
  (scim-enable-undo)
  (scim-disable-preedit-keymap))

(defun scim-cleanup-preedit ()
  (scim-remove-preedit)
  (if scim-debug (scim-message "cleanup preedit"))
  (setq scim-preedit-update nil
	scim-preedit-shown ""
	scim-preedit-string ""
	scim-preedit-prev-string ""
	scim-preedit-curpos 0
	scim-preedit-attributes nil
	scim-preedit-overlays nil
	scim-committed-string ""))

(defun scim-redraw-preedit ()
  (setq scim-preedit-update nil)
  (let ((str scim-preedit-string)
	(attrs scim-preedit-attributes))
    (if (or (string= scim-preedit-shown "FALSE")
	    (string= str ""))
	;; IMContext is empty or invisible
	(scim-cleanup-preedit)
      ;; IMContext cnatains preedit string
      (if (not (string= scim-preedit-prev-string ""))
	  (scim-remove-preedit)
	(scim-set-window-y-offset)
	(unless scim-surrounding-text-modified
	  (setq scim-preedit-default-attr nil)))
      ;; Put String
      (scim-enable-preedit-keymap)
      (scim-disable-undo)
      (setq scim-preedit-point (point))
;      (if scim-debug (scim-message "current cursor position: %d" scim-preedit-point))
      (insert str)
      (setq scim-preedit-prev-string str)
      ;; Set attributes
;      (if scim-debug (scim-message "attributes: %s" attrs))
      (let* ((max (length str))
	     (ol (make-overlay scim-preedit-point
			       (+ scim-preedit-point max)))
	     (flat-attr nil))
	(overlay-put ol 'face 'scim-preedit-default-face)
	(overlay-put ol 'priority 0)
	(setq scim-preedit-overlays (list ol))
	(while attrs
	  (let* ((beg (string-to-number (car attrs)))
		 (end (string-to-number
		       (car (setq attrs (cdr attrs)))))
		 (type (car (setq attrs (cdr attrs))))
		 (value (car (setq attrs (cdr attrs))))
		 fc pr)
;	    (if scim-debug (scim-message "beg: %d  end: %d  type: %s  val: %s" begin end type value))
	    (setq attrs (cdr attrs))
	    (if (cond ((and (string= type "foreground")
			    (scim-check-rgb-color value))
		       (setq fc (list :foreground value)
			     pr 50))
		      ((and (string= type "background")
			    (scim-check-rgb-color value))
		       (setq fc (list :background value)
			     pr 50))
		      ((and (string= type "decoreate")
			    (setq fc (scim-select-face-by-attr value)))
		       (setq pr 100)))
		(let ((ol (make-overlay (+ scim-preedit-point beg)
					(+ scim-preedit-point end))))
		  (overlay-put ol 'face fc)
		  (overlay-put ol 'priority pr)
		  (setq scim-preedit-overlays
			(cons ol scim-preedit-overlays))
		  (setq flat-attr (if (and (listp flat-attr)
					   (eq beg 0) (eq end max))
				      (cons fc flat-attr)
				    t)))
	      (scim-message "Unable to set attribute \"%s %s\"." type value))))
	(add-hook 'before-change-functions 'scim-before-change-function)
	(setq flat-attr (or flat-attr
			    'none)
	      scim-preedit-default-attr (or scim-preedit-default-attr
					    flat-attr))
	(if scim-debug (scim-message "default attr: %s" scim-preedit-default-attr))
	(if scim-debug (scim-message "current attr: %s" flat-attr))
	(if (or (eq flat-attr t)
		(not (equal flat-attr scim-preedit-default-attr)))
	    ;; When conversion candidate is shown
	    (scim-set-cursor-location)
	  ;; When the string is preedited or prediction window is drawn
	  (goto-char (+ scim-preedit-point scim-preedit-curpos))
	  (if (car scim-prediction-window-position)
	      (setq scim-preedit-curpos 0))
	  (if (cdr scim-prediction-window-position)
	      (scim-set-cursor-location)
	    (let ((scim-adjust-window-x-offset 0))
	      (scim-set-cursor-location))))
	))))

(defun scim-before-change-function (&optional beg end)
  (when (eq (current-buffer) scim-preedit-buffer)
    (if scim-debug (if beg (scim-message "change buffer (beg:%s  end:%s)" beg end)))
    (if scim-debug (if beg (scim-message "cursor positon: %s" (point))))
    (let ((str scim-preedit-string))
      (scim-cleanup-preedit)
      (unless scim-clear-preedit-when-unexpected-event
	(save-excursion
	  (goto-char scim-preedit-point)
	  (insert str)))
      (scim-reset-imcontext))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Manage switching to buffer
(defun scim-check-current-buffer ()
;  (if scim-debug (scim-message "check current buffer"))
  (let ((buffer (window-buffer)))
    (unless (eq buffer scim-last-command-buffer)
      ;; Focus out if buffer is switched to another
      (if scim-debug (scim-message "buffer was changed from %s to %s" scim-last-command-buffer buffer))
      (when scim-last-command-buffer
	(with-current-buffer scim-last-command-buffer
	  (when (stringp scim-imcontext-id)
	    (scim-change-focus nil)
	    (if scim-preedit-buffer
		;; Cleenup preedit if focus change become timeout
		(scim-before-change-function)))))
      (setq scim-last-command-buffer buffer)
      ;; Focus in if window is active
      (when (stringp scim-imcontext-id)
	(setq scim-frame-focus nil)
	(scim-check-frame-focus)))
    (if (or scim-mode-map-disabled buffer-read-only)
	;; Disable keymap
	(add-to-list 'minor-mode-overriding-map-alist '(scim-mode))
      ;; Enable keymap
      (setq minor-mode-overriding-map-alist
	    (delete '(scim-mode) minor-mode-overriding-map-alist)))
    ;; Switch IMContext between global and local
    (when (eq (local-variable-p 'scim-imcontext-id)
	      (not scim-mode-local))
      (scim-deregister-imcontext)
      (setq scim-last-command-buffer buffer))
    ;; Check whether the buffer has registered
    (unless scim-imcontext-id
      (if scim-debug (scim-message "new buffer was detected: %s" buffer))
      (scim-register-imcontext))))

(defun scim-kill-buffer-function ()
  (if (local-variable-p 'scim-imcontext-id)
      (scim-deregister-imcontext)
    (if (eq scim-last-command-buffer (current-buffer))
	(setq scim-last-command-buffer nil))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Communication with agent through an UNIX domain socket
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Disconnect
(defun scim-bridge-disconnect ()
  (let ((proc scim-bridge-socket))
    (setq scim-bridge-socket nil)
    (if (processp proc)
	(delete-process proc)
      (if scim-debug (scim-message "process is already deleted."))
      )))

(defun scim-bridge-process-sentinel (proc stat)
  (if scim-debug (scim-message "process: %s  status: %s" proc (substring stat 0 -1)))
  (scim-mode-quit))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Connect
(defun scim-bridge-connect-using-script ()
  (let* ((path scim-bridge-socket-script-path)
	 (script (concat path
			 (if (and (> (length path) 0)
				  (string= (substring path -1) "/"))
			     "" "/")
			 scim-bridge-socket-script-name)))
    (unless (and (file-exists-p script)
		 (file-executable-p script)
		 (save-current-buffer
		   (let* ((buf (find-file script))
			  (str (buffer-string)))
		     (kill-buffer buf)
		     (string= str scim-bridge-socket-script))))
      (unless (file-directory-p path)
	(make-directory path))
      (if scim-debug (scim-message "write the script file: %s" script))
      (with-temp-file script (insert scim-bridge-socket-script))
      (set-file-modes script ?\700))
    (start-process scim-bridge-name
		   scim-tmp-buffer-name
		   script
		   scim-bridge-socket-path)))

(defun scim-bridge-connect-internal ()
  (unless (file-exists-p scim-bridge-socket-path)
    (scim-message "launch scim-bridge...")
    (call-process-shell-command scim-bridge-name nil 0 nil "--noexit")
    (let ((i 30))
      (while (and (not (file-exists-p scim-bridge-socket-path))
		  (> (setq i (1- i)) 0))
	(sleep-for 0.1))
      (if (= i 0) (error "scim-bridge: %s" "socket not found!"))))
  (if (and (not scim-bridge-socket-use-script)
	   (featurep 'make-network-process))
      (make-network-process
       :name scim-bridge-name
       :service scim-bridge-socket-path
       :buffer scim-tmp-buffer-name
       :family 'local :server nil :noquery t)
    (scim-bridge-connect-using-script)))

(defun scim-bridge-connect ()
  (if (and (processp scim-bridge-socket)
	   (not (assq (process-status scim-bridge-socket) '(open run))))
      (scim-bridge-disconnect))
  (unless (and (processp scim-bridge-socket)
	       (assq (process-status scim-bridge-socket) '(open run)))
    (let ((proc (scim-bridge-connect-internal)))
      (if scim-debug (scim-message "process: %s  status: %s" proc (process-status proc)))
      (setq scim-bridge-socket proc)
      ;; `process-kill-without-query' is an obsolete function (as of Emacs 22.1)
;      (process-kill-without-query proc)
      (set-process-query-on-exit-flag proc nil)
      (set-process-coding-system proc 'utf-8 'utf-8)
      (set-process-sentinel proc 'scim-bridge-process-sentinel))
    (with-current-buffer scim-tmp-buffer-name
      (setq scim-tmp-buffer (current-buffer))
      (if scim-debug (scim-message "temp buffer: %s" scim-tmp-buffer))
      ;; `make-local-hook' is an obsolete function (as of Emacs 21.1)
;      (make-local-hook 'after-change-functions)
      (add-hook 'after-change-functions
		'scim-bridge-receive-commands nil t))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Communicate with agent
(defun scim-bridge-receive-commands (beg &optional end lng)
  (when (eq (current-buffer) scim-tmp-buffer)
    (accept-process-output nil 0 1)
    (let* ((i scim-bridge-timeout)
	   (repl (buffer-substring-no-properties beg (point-max))))
      (while (and (> (setq i (1- i)) 0)
		  (or (string= repl "")
		      (not (string= (substring repl -1) "\n"))))
	(accept-process-output nil 0 1)
	(setq repl (buffer-substring-no-properties beg (point-max))))
      (erase-buffer)
      (if scim-debug (scim-message "recv:\n%s" repl))
      (if scim-debug (scim-message "(received in %sms)" (- scim-bridge-timeout i)))
      (if (string= repl "")
	  (scim-message "Data reception became timeout.")
	(if scim-last-command-buffer
	    (with-current-buffer scim-last-command-buffer
	      (scim-parse-reply (scim-split-commands repl)))))
      )))

(defun scim-bridge-send-command (cmd)
  (with-current-buffer scim-tmp-buffer
    (remove-hook 'after-change-functions 'scim-bridge-receive-commands t)
    (erase-buffer)
    (if scim-debug (scim-message "process: %s  status: %s" scim-bridge-socket (process-status scim-bridge-socket)))
    (if scim-debug (scim-message "send: %S" cmd))
    (unwind-protect
	(progn
	  (process-send-string scim-bridge-socket (concat cmd "\n"))
	  (if scim-bridge-wait-reply (scim-bridge-receive-commands 1)))
      (add-hook 'after-change-functions
		'scim-bridge-receive-commands nil t))))

(defun scim-bridge-send-only (cmd)
  (let ((scim-bridge-wait-reply nil))
    (scim-bridge-send-command cmd)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Send command to agent
(defun scim-register-imcontext ()
  (unless scim-imcontext-id
    (when scim-mode-local
      (make-local-variable 'scim-imcontext-id)
      (put 'scim-imcontext-id 'permanent-local t))
    (setq scim-imcontext-id 'RQ)
    (scim-bridge-send-command "register_imcontext")))

(defun scim-deregister-imcontext () ;(id)
  (if (stringp scim-imcontext-id)
      (progn
	(scim-change-focus nil)
	(scim-bridge-send-command
	 (concat "deregister_imcontext " scim-imcontext-id)))
    (if (local-variable-p 'scim-imcontext-id)
	(kill-local-variable 'scim-imcontext-id)
      (setq-default scim-imcontext-id nil))
    (if (eq scim-last-command-buffer (current-buffer))
	(setq scim-last-command-buffer nil))))

(defun scim-reset-imcontext () ;(id)
;  (if scim-debug (scim-message "buffer: %s" (current-buffer)))
  (scim-bridge-send-command
   (concat "reset_imcontext " scim-imcontext-id)))

(defun scim-set-preedit-mode () ;(id mode)
  (scim-bridge-send-command
   (concat "set_preedit_mode " scim-imcontext-id " embedded")))

(defun scim-change-focus (focus-in) ;(id focus-in)
  (scim-bridge-send-command
   (concat "change_focus " scim-imcontext-id
	   (if focus-in " true" " false"))))

(defun scim-set-cursor-location () ;(id x y)
  (let* ((pixpos (save-excursion
		   (goto-char (+ scim-preedit-point scim-preedit-curpos))
		   (scim-compute-pixel-position)))
	 (x (number-to-string
	     (max (- (car pixpos) scim-adjust-window-x-offset) 1)))
	 (y (number-to-string (cdr pixpos))))
    (if scim-debug (scim-message "cursor position: (%s, %s)" x y))
    (scim-bridge-send-only
     (concat "set_cursor_location " scim-imcontext-id " " x " " y))))

(defun scim-handle-key-event (key-code key-pressed modifiers) ;(id key-code key-pressed &rest modifiers)
  (scim-bridge-send-command
   (scim-construct-command
    (append (list "handle_key_event"
		  scim-imcontext-id key-code key-pressed)
	    modifiers))))

(defun scim-preedit-updated ()
  (scim-bridge-send-only "preedit_updated"))

(defun scim-string-commited ()
  (scim-bridge-send-only "string_commited"))

(defun scim-surrounding-text-gotten (retval cursor-position string)
  (scim-bridge-send-command
   (scim-construct-command
    (cons "surrounding_text_gotten"
	  (if retval
	      (list "true" (number-to-string cursor-position) string)
	    (list "false"))))))

(defun scim-surrounding-text-deleted (retval)
  (scim-bridge-send-command
   (concat "surrounding_text_deleted " (if retval "true" "false"))))

(defun scim-surrounding-text-replaced (retval)
  (scim-bridge-send-command
   (concat "surrounding_text_replaced " (if retval "true" "false"))))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Process commands from agent to clients
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun scim-imengine-status-changed (id enabled)
  (scim-cleanup-preedit))

(defun scim-preedit-mode-changed ()
  t)

(defun scim-imcontext-registered (id)
  ;; Initialize IMContext
  (if scim-debug (scim-message "imcontext registered (id: %s  buf: %s)" id (if scim-mode-local (current-buffer) "global")))
  (if scim-focus-update-timer (cancel-timer scim-focus-update-timer))
  (setq scim-imcontext-id id
	scim-preedit-point (point)
	scim-preedit-prev-string ""
	scim-preedit-overlays nil)
  (scim-cleanup-preedit)
  (scim-set-preedit-mode)
  (setq scim-frame-focus nil)
  (scim-check-frame-focus)
  (setq scim-focus-update-timer
	(run-at-time 0 scim-focus-update-interval 'scim-check-frame-focus))
  (scim-set-cursor-location))

(defun scim-imcontext-deregister ()
  (if scim-debug (scim-message "imcontext deregistered (id: %s  buf: %s)" scim-imcontext-id (if (local-variable-p 'scim-imcontext-id) (current-buffer) "global")))
  (if (local-variable-p 'scim-imcontext-id)
      (kill-local-variable 'scim-imcontext-id)
    (setq-default scim-imcontext-id nil))
  (setq scim-last-command-buffer nil))

(defun scim-imcontext-reseted (id)
  t)

(defun scim-cursor-location-changed ()
  t)

(defun scim-key-event-handled (consumed)
  (if (or (string= consumed "true")
	  (not scim-last-command-event))
      ;; If key event is handled
      (setq scim-last-command-event nil)
    ;; If key event is ignored
    (let* ((vec (vector scim-last-command-event))
	   (event (or (and (boundp 'local-function-key-map)
			   (lookup-key local-function-key-map vec))
		      (lookup-key function-key-map vec)))
	   keybind)
      (setq event (or (and (arrayp event)
			   (car (listify-key-sequence event)))
		      scim-last-command-event))
      (let ((minor-mode-overriding-map-alist
	     (cons '(scim-mode) minor-mode-overriding-map-alist))
	    (scim-mode-map-alist nil))
	(setq keybind (key-binding (vector event)))
	(if (and (not keybind)
		 (integerp event))
	    ;; Reset the 25th bit corresponding to the shift key
	    (setq event (logand event (lognot ?\x2000000))
		  keybind (key-binding (vector event)))))
      (if scim-debug (scim-message "event: --> %s --> %s" scim-last-command-event event))
      (setq scim-last-command-event nil
	    last-command-event event
	    this-command keybind)
      (if scim-debug (scim-message "execute command: %s" keybind))
      (cond ((not (commandp keybind))
	     (scim-message "%s is undefined"
			   (single-key-description (aref vec 0))))
	    ((eq keybind 'self-insert-command)
	     (scim-insert-and-modify-undo-list (char-to-string event)))
	    (t (command-execute keybind))))
      ))

(defun scim-update-preedit (id)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (setq scim-preedit-update t)))

(defun scim-set-preedit-string (id string)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (setq scim-preedit-string string)))

(defun scim-set-preedit-attributes (id &rest attrs)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (setq scim-preedit-attributes attrs)))

(defun scim-set-preedit-cursor-position (id position)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (if scim-debug (scim-message "cursor position: %s" position))
    (setq scim-preedit-curpos (string-to-number position))))

(defun scim-set-preedit-shown (id shown)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (setq scim-preedit-shown shown)))

(defun scim-set-commit-string (id string)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (if scim-debug (scim-message "commit string: %s" string))
    (setq scim-committed-string string)))

(defun scim-commit-string (id)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (let ((str scim-committed-string))
      (scim-remove-preedit)
      (if scim-undo-by-committed-string
	  (insert str)
	(scim-insert-and-modify-undo-list str))
      (unless this-command (undo-boundary))
      (setq scim-preedit-prev-string ""
	    scim-surrounding-text-modified t)
      (scim-redraw-preedit))))

(defun scim-forward-key-event (id key-code key-pressed &rest modifiers)
  (let ((event (scim-encode-event key-code modifiers)))
    (if event
	(if (string= key-pressed "TRUE")
	    (setq unread-command-events (cons event
					      unread-command-events)))
      (if (not (string= id scim-imcontext-id))
	  (scim-message "IMContext ID (%s) is mismatched." id)
	(scim-handle-key-event key-code key-pressed modifiers))
      )))

(defun scim-get-surrounding-text (id before-max after-max)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (let* ((len-before (scim-twos-complement before-max))
	   (len-after (scim-twos-complement after-max))
	   (end-before (if scim-preedit-buffer
			   scim-preedit-point
			 (point)))
	   (beg-after (if scim-preedit-buffer
			  (+ end-before (length scim-preedit-prev-string))
			end-before))
	   (beg-before (if (>= len-before 0)
			   (max (- end-before len-before) (point-min))
			 (line-beginning-position)))
	   (end-after (if (>= len-after 0)
			  (min (+ beg-after len-after) (point-max))
			(line-end-position)))
	   (str-before (buffer-substring-no-properties beg-before end-before))
	   (str-after (buffer-substring-no-properties beg-after end-after))
	   (string (concat str-before str-after))
	   (cursor (length str-before))
	   (retval (> (length string) 0)))
      (if scim-debug (scim-message "val: %s  str: %s  pos: %d" retval string cursor))
      (scim-surrounding-text-gotten retval cursor string))))

(defun scim-delete-surrounding-text (id offset length)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (if scim-debug (scim-message "delete surrounding text"))
    (scim-remove-preedit)
    (let* ((pos (point))
	   (beg (+ pos (scim-twos-complement offset)))
	   (end (+ beg (string-to-number length))))
      (delete-region beg end))
    (setq scim-preedit-prev-string ""
	  scim-surrounding-text-modified t)
    (scim-redraw-preedit)
    (scim-surrounding-text-deleted t)))

(defun scim-replace-surrounding-text (id corsor-index string)
  (if (not (string= id scim-imcontext-id))
      (scim-message "IMContext ID (%s) is mismatched." id)
    (if scim-debug (scim-message "replace surrounding text"))
    (scim-remove-preedit)
    (let* ((pos (point))
	   (beg (- pos (scim-twos-complement corsor-index)))
	   (end (+ beg (length string))))
      (delete-region beg end)
      (goto-char beg)
      (insert string)
      (goto-char pos))
    (setq scim-preedit-prev-string ""
	  scim-surrounding-text-modified t)
    (scim-redraw-preedit)
    (scim-surrounding-text-replaced t)))

(defun scim-focus-changed ()
  t)

(defun scim-beep (id)
  (ding t))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Execute commands replied from agent
(defun scim-parse-reply (cmdlist)
  (if scim-debug (scim-message "this-command: %s" this-command))
  (while cmdlist
    (let* ((args (car cmdlist))
	   (cmd (cdr (assoc (car args) scim-reply-alist))))
      (if cmd
	  (progn
	    (if scim-debug (scim-message "execute: %s" args))
	    (apply cmd (cdr args)))
	(scim-message "Unknown command received from agent: %s" (symbol-name cmd))
	))
    (setq cmdlist (cdr cmdlist)))
  (when scim-preedit-update
    (if scim-debug (scim-message "preedit-update  win-buf: %s  cur-buf: %s  cmd-buf: %s  str: \"%s\"" (window-buffer) (current-buffer) scim-preedit-buffer scim-preedit-string))
    (scim-redraw-preedit)
    (scim-preedit-updated)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Process key events
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun scim-wait-following-key-event (prev-event key-code modifiers)
  (let ((event (read-event nil nil scim-key-release-delay)))
    (if (eq event prev-event)
	(setq unread-command-events (cons event
					  unread-command-events))
      (let ((scim-key-release-delay nil))
	(scim-dispatch-key-event event)))
    (scim-handle-key-event key-code "false" modifiers)))

(defun scim-dispatch-key-event (event)
  (let* ((elist (scim-decode-event event))
	 (key-code (car elist))
	 (modifiers (cdr elist)))
    (if (numberp key-code) (scim-check-current-buffer))
;    (if scim-debug (scim-message "event: %s" elist))
    (if scim-debug (scim-message "event: %s" event))
    (setq scim-last-command-event event
	  scim-surrounding-text-modified nil)
    (if (and (stringp scim-imcontext-id)
	     (numberp key-code))
	(progn
	  ;; Send a key event to agent
	  (setq key-code (number-to-string key-code))
	  (scim-handle-key-event key-code "true" modifiers)
	  (if scim-key-release-delay
	      (scim-wait-following-key-event event key-code modifiers)
	    (scim-handle-key-event key-code "false" modifiers)))
      ;; IMContext is not registered or key event is not recognized
      (scim-key-event-handled "false")))
  (unless (memq 'scim-check-current-buffer post-command-hook)
    (if scim-debug (scim-message "`post-command-hook' was reset. now add the hook again."))
    (add-hook 'post-command-hook 'scim-check-current-buffer)))

(defun scim-handle-event (&optional arg)
  (interactive "*p")
  (scim-dispatch-key-event last-command-event))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; Switch minor mode
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
(defun scim-mode-on ()
  (if (not window-system)
      (scim-mode-quit)
    (if scim-bridge-socket (scim-mode-off)) ;; Restart scim-mode
    (unwind-protect
	(scim-bridge-connect)
      (if (not (and scim-bridge-socket
		    (memq (process-status scim-bridge-socket)
			  '(open run))))
	  ;; Connection failed
	  (scim-mode-quit)
	;; Connection was established
	;; Cleenup variables
	(let ((buffers (buffer-list)))
	  (save-current-buffer
	    (while buffers
	      (set-buffer (car buffers))
	      (kill-local-variable 'scim-imcontext-id)
	      (setq buffers (cdr buffers)))))
	(setq-default scim-imcontext-id nil)
	(setq scim-last-command-buffer nil
	      scim-preedit-default-attr nil)
	;; Initialize key binding
	(scim-update-key-bindings)
	(when (and (fboundp 'add-to-ordered-list)
		   (boundp 'emulation-mode-map-alists))
	  (add-to-ordered-list
	   'emulation-mode-map-alists 'scim-mode-map-alist 50))
	;; Turn on minor mode
	(mapc (lambda (hook)
		(add-hook hook 'scim-disable-keymap))
	      scim-incompatible-mode-hooks)
	(add-hook 'post-command-hook 'scim-check-current-buffer)
	(if scim-debug (scim-message "post-command-hook: %s" post-command-hook))
	(add-hook 'kill-buffer-hook 'scim-kill-buffer-function)
	(add-hook 'kill-emacs-hook 'scim-mode-off)
	(setq-default scim-mode t)))
    (scim-update-mode-line)))

(defun scim-mode-quit ()
  (remove-hook 'kill-emacs-hook 'scim-mode-off)
  (remove-hook 'kill-buffer-hook 'scim-kill-buffer-function)
  (remove-hook 'post-command-hook 'scim-check-current-buffer)
  (mapc (lambda (hook)
	  (remove-hook hook 'scim-disable-keymap))
	scim-incompatible-mode-hooks)
  (kill-buffer scim-tmp-buffer)
  (if scim-focus-update-timer (cancel-timer scim-focus-update-timer))
  (scim-cleanup-preedit)
  (when (and (fboundp 'add-to-ordered-list)
	     (boundp 'emulation-mode-map-alists))
    (setq emulation-mode-map-alists
	  (delq 'scim-mode-map-alist emulation-mode-map-alists)))
  (scim-bridge-disconnect)
  (setq-default scim-mode nil)
  (scim-update-mode-line))

(defun scim-mode-off ()
  ;; Deregister IMContext IDs
  (let ((buffers (buffer-list)))
    (save-current-buffer
      (while buffers
	(set-buffer (car buffers))
	(when (local-variable-p 'scim-imcontext-id)
	  (setq scim-last-command-buffer (current-buffer))
	  (scim-deregister-imcontext))
	(setq buffers (cdr buffers)))))
  (scim-deregister-imcontext)
  ;; Turn off minor mode
  (scim-mode-quit))

(defun scim-update-mode ()
  (if scim-mode
      (scim-mode-on)
    (scim-mode-off)))

(defun scim-mode (&optional arg)
  "Start/Stop Scim conversion system."
  (interactive "P")
  (if (not (or window-system scim-mode))
      (prog1 nil
	(scim-message "scim-mode needs Emacs to run as X application."))
    (setq-default scim-mode
		  (if (null arg)
		      (not scim-mode)
		    (> (prefix-numeric-value arg) 0)))
    (scim-update-mode)))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; Setup minor-mode
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; minor-mode-alist
(unless (assq 'scim-mode minor-mode-alist)
    (setq minor-mode-alist
       (cons '(scim-mode scim-mode-line-string)
	     minor-mode-alist)))
; minor-mode-map-alist
(unless (assq 'scim-mode minor-mode-map-alist)
    (setq minor-mode-map-alist
       (cons (cons 'scim-mode scim-mode-map)
	     minor-mode-map-alist)))

(provide 'scim-bridge)

;;;
;;; scim-bridge.el ends here
